/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("jazz.app.auth");

dojo.require("jazz.app.internal.AuthenticationInterceptor");
dojo.require("jazz.app.internal.HeadlessCredentialProvider");
dojo.require("jazz.app.internal.LoginDialog");
dojo.require("jazz.app.internal.OAuthInterceptor");
dojo.require("jazz.client.xhr");

var loginDialog = new jazz.app.internal.LoginDialog();
jazz.app.internal.AuthenticationInterceptor.setCredentialProvider(loginDialog);

// Apply Jazz authentication logic to responses received through jazz.client.xhr
jazz.client.internal.addResponseInterceptor(jazz.app.internal.AuthenticationInterceptor);

// Apply OAuth interceptor logic
jazz.client.internal.addResponseInterceptor(jazz.app.internal.OAuthInterceptor);

/**
 * Attempts to log in to the local server using either the supplied user ID and
 * password, or by prompting the user for credentials. Note: This function will
 * first call the log out method before attempting to log in.
 * </p><p>
 * This function takes two forms:
 * </p><ul>
 * <li>Zero argument form: Prompts the user to enter credentials.</li>
 * <li>One argument form: If the single argument is an object containing a
 *   property 'userId' with a string value and a property 'password' with a
 *   string value, attempts to log in with those credentials. If either of these
 *   properties are missing or contain a non-string value, uses the zero
 *   argument behavior (prompts for credentials).</li>
 * </ul><p>
 * On a successful log in, the system calls the Deferred's callback chain with
 * an unspecified return type; clients may only assume that the log in attempt
 * succeeded based on the fact that the callback chain was called. Note that a
 * successful log in also fires the jazz.app.events.LOGIN event.
 * </p><p>
 * On an unsuccessful log in attempt, the system calls the Deferred's errback
 * chain with an Error object containing information about the error that
 * occurred.
 * </p><p>
 * Note: The system will automatically challenge a user for credentials as long
 * as some client has dojo.require'd this module. It is unnecessary to manually
 * call log in unless you want to invoke an explicit log in (e.g. when a user
 * clicks on a "Log in" command).
 * </p>
 * 
 * @return {dojo.Deferred} a dojo.Deferred object indicating success or failure
 *  of the log out attempt.
 * @since 0.7
 */
jazz.app.auth.login = function(args) {
	
	var logoutDeferred = jazz.app.auth.logout();
	
	var clientDeferred = new dojo.Deferred(function(deferred) {
		logoutDeferred.cancel();
	});
	
	logoutDeferred.addErrback(function(logoutError) {
		clientDeferred.errback(logoutError);
	});
	
	logoutDeferred.addCallback(function(logoutResult) {
		if(dojo.isObject(args) && args.userId && args.password) {
			// if the user specified credentials, switch the auth interceptor to use
			// those credentials rather than prompting the user
			var headlessCredentialProvider = new jazz.app.internal.HeadlessCredentialProvider(args.userId, args.password);
			jazz.app.internal.AuthenticationInterceptor.setCredentialProvider(headlessCredentialProvider);
		}
		
		var dfd = jazz.client.xhrPost({
			// TODO: Explore using a standard URI to force log in
			url: net.jazz.ajax._contextRoot + "/web/loggedInUser"
		});
		
		dfd.addBoth(function(result) {
			if(headlessCredentialProvider) {
				// only use a headless provider once
				jazz.app.internal.AuthenticationInterceptor.setCredentialProvider(loginDialog);
			}
			return result;
		});
		
		dfd.addCallback(function(result) {
			// hide response body
			clientDeferred.callback(null);
		});
		
		dfd.addErrback(function(loginError) {
			clientDeferred.errback(loginError);
		});
	});
	
	return clientDeferred;
};

/**
 * Causes a log in challenge to occur if the user is not already logged in. This
 * is useful because jazz.app.auth.login always first calls jazz.app.auth.logout
 * which may not always be desirable to the caller.
 * 
 * @return {dojo.Deferred} a dojo.Deferred object indicating success or failure
 *  of the log in attempt.
 * @since 0.7
 */
jazz.app.auth.loginIfNecessary = function() {
	var loginDeferred = jazz.client.xhrPost({
		// TODO: Explore using a standard URI to force log in
		url: net.jazz.ajax._contextRoot + "/web/loggedInUser"
	});
	
	var clientDeferred = new dojo.Deferred(function(deferred) {
		loginDeferred.cancel();
	});
	
	loginDeferred.addCallback(function(result) {
		// hide response body
		clientDeferred.callback(null);
	});
	
	loginDeferred.addErrback(function(loginError) {
		clientDeferred.errback(loginError);
	});
	
	return clientDeferred;
};

/**
 * Attempts to log out of the local server.
 * <p>
 * On a successful log out, the system calls the Deferred's callback chain with
 * an unspecified return type; clients may only assume that the log in attempt
 * succeeded based on the fact that the callback chain was called. Note that a
 * successful log in also fires the jazz.app.events.LOGOUT event.
 * </p><p>
 * On an unsuccessful log out attempt, the system calls the Deferred's errback
 * chain with an Error object containing information about the error that
 * occurred.
 * </p>
 * 
 * @return {dojo.Deferred} a dojo.Deferred object indicating success or failure
 *  of the log out attempt.
 * @since 0.7
 */
jazz.app.auth.logout = function() {
	var dfd = jazz.client.xhrPost({
		url: net.jazz.ajax._contextRoot + "/auth/logout"
	});
	
	dfd.addCallback(function(result) {
		// hide response body
		return null;
	});
	
	return dfd;
};
})();